import { useEffect, useMemo, useRef, useState } from "react";
import { EventArgs, Graph, Shape, Point } from "@antv/x6";
import { Transform } from "@antv/x6-plugin-transform";
import { Scroller } from "@antv/x6-plugin-scroller";
import { Snapline } from "@antv/x6-plugin-snapline";
import { Keyboard } from "@antv/x6-plugin-keyboard";
import { Export } from "@antv/x6-plugin-export";
import { GetAllDesignTables } from "@/api";
import { useRequest } from "umi";
import type {
  ColumnItem,
  ColumnRelation,
  ProjectInfo,
  RemarkInfo,
  TableItemType,
  TopicAreaInfo,
} from "@/type";
import { uuid } from "@/utils";
import { RelationType } from "@/enum";
import { render } from "./renderer";

import "@/components/TableNode";
import "@/components/TopicNode";
import "@/components/NoticeNode";
import { message } from "antd";

export default function erModel() {
  const graphRef = useRef<Graph>();
  const [graph, setGraph] = useState<Graph>();
  const historyRef = useRef<ProjectInfo[]>([]);
  const activeIndex = useRef(0);
  const [project, setProjectInfo] = useState<ProjectInfo>({
    id: "1",
    name: "项目1",
    folderId: "root",
    creationTime: "2024-12-12 12:12:12",
    creatorUser: "管理员",
    lastModificationTime: "2024-12-12 12:12:12",
    description: "模型描述",
    isTemplate: false,
    industry: "互联网",
    publishStatus: "未发布",
    tables: [],
    relations: [],
    topicAreas: [],
    remarks: [],
    history: [],
    setting: {
      showMenu: true,
      showSidebar: true,
      showColumnDetail: true,
      showGrid: true,
      showRelation: true,
    },
  });

  /**
   * 统一修改数据
   * @param info 模型信息
   * @param ingoreHistory 忽略历史记录
   * @param isInit  初始化
   */
  const setProject = (
    info: ProjectInfo | ((state: ProjectInfo) => ProjectInfo),
    ingoreHistory?: boolean,
    isInit?: boolean
  ) => {
    if (isInit) {
      historyRef.current = [];
      activeIndex.current = 0;
    }

    if (info && typeof info === "function") {
      setProjectInfo((state) => {
        const result = info(state);
        graphRef.current && render(graphRef.current, result);
        // 添加记录
        if (!ingoreHistory) {
          historyRef.current?.push(result);
          activeIndex.current = historyRef.current?.length - 1;
          if (historyRef.current?.length > 20) {
            historyRef.current?.shift();
            activeIndex.current -= 1;
          }
        }
        return result;
      });
    } else {
      setProjectInfo(info);
      graphRef.current && render(graphRef.current, info);
      // 添加记录
      if (!ingoreHistory) {
        historyRef.current?.push(info);
        activeIndex.current = historyRef.current?.length - 1;
        if (historyRef.current?.length > 20) {
          historyRef.current?.shift();
          activeIndex.current -= 1;
        }
      }
    }
  };

  const initGraph = (container: HTMLElement) => {
    const instance = new Graph({
      container,
      width: document.documentElement.clientWidth,
      height: document.documentElement.clientHeight,
      autoResize: true,
      async: false,
      mousewheel: {
        enabled: true,
        modifiers: "ctrl",
        minScale: 0.2,
        maxScale: 2,
      },
      highlighting: {
        nodeAvailable: {
          name: "stroke",
          args: {
            padding: 4,
            attrs: {
              "stroke-width": 2,
              stroke: "red",
            },
          },
        },
      },
      connecting: {
        allowBlank: false,
        allowEdge: false,
        allowLoop: false,
        router: {
          name: "normal",
        },
        createEdge() {
          return new Shape.Edge({
            attrs: {
              line: {
                stroke: "#ff0000",
                strokeWidth: 1,
                strokeDasharray: 5,
                targetMarker: null,
              },
            },
            data: {
              type: "refer",
            },
          });
        },
      },
      grid: {
        visible: true,
        size: 10,
      },
      background: {
        color: "#F2F7FA",
      },
      interacting: {
        nodeMovable: (view) => {
          const data = view.cell.getData<{
            ignoreDrag: boolean;
            lock: boolean;
          }>();
          // 禁止拖拽或锁节点
          if (data?.ignoreDrag || data?.lock) return false;
          return true;
        },
      },
    });

    instance.use(new Snapline({ enabled: true }));
    instance.use(
      new Transform({
        resizing: {
          enabled: (node) => {
            return node.shape !== "table-node";
          },
        },
      })
    );
    instance.use(new Scroller());
    instance.use(new Keyboard());
    instance.use(new Export());

    setGraph(instance);
    graphRef.current = instance;

    instance.on(
      "node:change:add:relation",
      (args: EventArgs["cell:change:*"]) => {
        console.log("node:change:add:relation", args.current);
        const { source, target } = args.current;
        if (source && target) {
          addRelation(source, target);
        }
      }
    );

    instance.on(
      "node:change:update:remark",
      function (args: EventArgs["cell:change:*"]) {
        console.log('修改备注：', args.current)
        updateRemark(args.current);
      }
    );

    // Graph.registerAnchor({
    //   'custom-anchor': function(view, magnet, ref, options) {
    //     // console.log(view, magnet, ref);
    //     const source = magnet.getBoundingClientRect();
    //     const target = (ref as SVGAElement).getBoundingClientRect();
    //     console.log(source, target, this);
    //     if(source.x + source.width < target.x + target.width) {
    //       // 右侧
    //       // 获取右侧点
    //       if(this.sourceView?.cell.isNode()) {
    //         return new Point(this.sourceView.getBBox().x + this.sourceView.getBBox().width, this.sourceView.getBBox().y + this.sourceView.getBBox().height / 2);
    //       }
    //     } else {
    //       // 左侧
    //       return new Point(this.sourceView.getBBox().x, this.sourceView.getBBox().y + this.sourceView.getBBox().height / 2);
    //     }
    //     return new Point(100, 200);
    //   }
    // });
  };

  // 能否重做
  const canRedo = useMemo(() => {
    return (
      historyRef.current?.length > 1 &&
      activeIndex.current < historyRef.current?.length - 1
    );
  }, [historyRef.current, activeIndex.current]);

  // 能否撤销
  const canUndo = useMemo(() => {
    return activeIndex.current > 0 && historyRef.current?.length > 1;
  }, [historyRef.current, activeIndex.current]);

  // 撤销
  const onUndo = () => {
    const info = historyRef.current?.[activeIndex.current - 1];
    activeIndex.current -= 1;
    setProject(info, true);
  };

  // 重做
  const onRedo = () => {
    const info = historyRef.current?.[activeIndex.current + 1];
    activeIndex.current += 1;
    setProject(info, true);
  };

  /**
   * 添加表
   */
  const addTable = (parentId?: string) => {
    const tableId = uuid();
    const columnId = uuid();
    const newTable: TableItemType = {
      table: {
        aliasName: "newtable",
        creationTime: "",
        creatorUserId: "",
        displayOrder: 1,
        id: tableId,
        isDeleted: false,
        langDescription: "",
        langName: "",
        name: "",
        cn_name: "新建表",
        en_name: "new table",
        parentBusinessTableId: parentId || "",
        schemaName: "new_table",
        type: 1,
        updateTime: "",
        openSync: false,
        style: {
          // 随机颜色
          color: "#" + Math.floor(Math.random() * 0x666666).toString(16),
          x: 0,
          y: 0,
        },
      },
      tableColumnList: [
        {
          aggregateEnable: false,
          businessTableId: tableId,
          charset: "",
          defaultValue: "",
          displayEnable: false,
          displayOrder: 1,
          groupByEnable: false,
          id: columnId,
          isAggregateEnable: false,
          isDisplayEnable: false,
          isGroupByEnable: false,
          isLinkEnable: false,
          isOrderByEnable: false,
          isPreDefined: true,
          isRequired: true,
          isUnique: false,
          isWhereEnable: false,
          langName: "120d5c4e-a5d8-4d49-b5f5-492feb37e6c6",
          linkEnable: false,
          maxLength: 50,
          name: "Id",
          orderByEnable: false,
          preDefined: true,
          precision: 0,
          required: true,
          scale: 0,
          schemaName: "id",
          type: 2,
          unique: false,
          whereEnable: false,
          cn_name: "id",
          en_name: "id",
        },
      ],
    };

    // 子表插入到父表后面
    const tables = [...project.tables];
    tables.splice(
      tables.findIndex((item) => item.table.id === parentId) + 1,
      0,
      newTable
    );
    setProject({
      ...project,
      tables: parentId ? tables : [...project.tables, newTable],
    });
  };

  /**
   * 更新表
   * @param table
   */
  const updateTable = (table: TableItemType) => {
    setProject({
      ...project,
      tables: project.tables.map((item) => {
        if (item.table.id === table.table.id) {
          const tableNode = graphRef.current?.getCellById(table.table.id);
          tableNode?.setData(table);
          return table;
        }
        return item;
      }),
    });
  };

  /**
   * 删除表及其子表
   * @param tableId
   */
  const deleteTable = (tableId: string) => {
    setProject({
      ...project,
      tables: project.tables.filter(
        (item) =>
          item.table.id !== tableId ||
          item.table.parentBusinessTableId !== tableId
      ),
      // todo 处理子表对应关系
      relations: project.relations.filter(
        (item) => item.primaryTable !== tableId && item.foreignTable !== tableId
      ),
    });
    graphRef.current?.removeCell(tableId);
    // 删除关联关系
    project.relations.forEach((item) => {
      if (item.primaryTable === tableId || item.foreignTable === tableId) {
        graphRef.current?.removeCell(item.id);
      }
    });
  };

  /**
   * 增加主题域
   */
  const addTopicArea = () => {
    const topicAreaId = uuid();
    const newTopicArea = {
      id: topicAreaId,
      name: "主题域_" + (project.topicAreas.length + 1),
      style: {
        background: "#175e7a",
      },
    };
    setProject({
      ...project,
      topicAreas: [...project.topicAreas, newTopicArea],
    });
  };

  /**
   * 修改主题域
   */
  const updateTopicArea = (topicArea: TopicAreaInfo) => {
    setProject({
      ...project,
      topicAreas: project.topicAreas.map((item) => {
        if (item.id === topicArea.id) {
          const topicAreaNode = graphRef.current?.getCellById(topicArea.id);
          topicAreaNode?.setData(topicArea);
          return topicArea;
        }
        return item;
      }),
    });
  };

  /**
   * 删除主题域
   * @param topicAreaId
   */
  const deleteTopicArea = (topicAreaId: string) => {
    setProject({
      ...project,
      topicAreas: project.topicAreas.filter((item) => item.id !== topicAreaId),
    });
    graphRef.current?.removeCell(topicAreaId);
  };

  /**
   * 添加备注
   */
  const addRemark = () => {
    const remarkId = uuid();
    const newRemark = {
      id: remarkId,
      name: "备注_" + (project.remarks.length + 1),
      text: "",
      style: {
        x: 0,
        y: 0,
        width: 200,
        height: 100,
        background: "#fcf7ac",
      },
    };
    setProject({
      ...project,
      remarks: [...project.remarks, newRemark],
    });
  };

  /**
   * 修改备注
   */
  const updateRemark = (remark: RemarkInfo) => {
    setProject((state) => ({
      ...(state || {}),
      remarks: state.remarks.map((item) => {
        if (item.id === remark.id) {
          const remarkNode = graphRef.current?.getCellById(remark.id);
          remarkNode?.setData(remark);
          return remark;
        }
        return item;
      }),
    }));
  };

  /**
   * 删除备注
   * @param remarkId
   */
  const deleteRemark = (remarkId: string) => {
    setProject({
      ...project,
      remarks: project.remarks.filter((item) => item.id !== remarkId),
    });
    graphRef.current?.removeCell(remarkId);
  };

  const getRelations = (project: ProjectInfo, newRelation: ColumnRelation) => {
    let sourceColumn: ColumnItem | undefined;
    let targetColumn: ColumnItem | undefined;
    let sourceTable: TableItemType | undefined;
    let targetTable: TableItemType | undefined;
    project.tables.forEach((table) => {
      if (table.table.id === newRelation.primaryTable) {
        sourceTable = table;
        sourceColumn = table.tableColumnList.find(
          (item) => item.id === newRelation.primaryKey
        );
      }
      if (table.table.id === newRelation.foreignTable) {
        targetTable = table;
        targetColumn = table.tableColumnList.find(
          (item) => item.id === newRelation.foreignKey
        );
      }
    });

    if (!sourceColumn || !targetColumn) {
      return {
        relations: project.relations,
        canAdd: false,
      };
    }
    if (sourceColumn.type !== targetColumn.type) {
      message.warning("数据类型不一致");
      return {
        relations: project.relations,
        canAdd: false,
      };
    }

    return {
      relations: [
        ...project.relations,
        {
          ...newRelation,
          name: `${sourceTable?.table.schemaName}_${targetTable?.table.schemaName}_${sourceColumn.schemaName}`,
        },
      ],
      canAdd: true,
    };
  };

  /**
   * 添加关系
   */
  const addRelation = (
    source: {
      tableId: string;
      columnId: string;
    },
    target: {
      tableId: string;
      columnId: string;
    }
  ) => {
    const newRelation: ColumnRelation = {
      id: uuid(),
      name: "",
      primaryKey: source.columnId,
      primaryTable: source.tableId,
      foreignKey: target.columnId,
      foreignTable: target.tableId,
      relationType: 1,
      style: {},
    };
    setProject((state) => {
      const obj = getRelations(state, newRelation);
      if (obj.canAdd) {
        return {
          ...state,
          relations: obj.relations,
        };
      } else {
        return state;
      }
    });
  };

  /**
   * 更新关系
   */
  const updateRelation = (relation: ColumnRelation) => {
    setProject((state) => {
      return {
        ...state,
        relations: state.relations.map((item) => {
          if (item.id === relation.id) {
            return relation;
          }
          return item;
        }),
      };
    });
    // 更新连线
    const relationEdge = graphRef.current?.getCellById(relation.id);
    if (relationEdge?.isEdge()) {
      relationEdge.setSource({
        cell: relation.primaryTable,
        port: relation.primaryKey + "_port2",
        anchor: "left",
      });
      relationEdge.setTarget({
        cell: relation.foreignTable,
        port: relation.foreignKey + "_port2",
        anchor: "left",
      });
      // 更新标签
      relationEdge.setLabelAt(0, {
        attrs: {
          txt: {
            text:
              relation.relationType === RelationType.OneToOne ||
              relation.relationType === RelationType.OneToMany
                ? "1"
                : "n",
          },
        },
        position: {
          distance: 25,
        },
      });
      relationEdge.setLabelAt(1, {
        attrs: {
          txt: {
            text:
              relation.relationType === RelationType.OneToMany ||
              relation.relationType === RelationType.ManyToMany
                ? "n"
                : "1",
          },
        },
        position: {
          distance: -25,
        },
      });
    }
  };

  /**
   * 删除关系
   */
  const deleteRelation = (relationId: string) => {
    setProject({
      ...project,
      relations: project.relations.filter((item) => item.id !== relationId),
    });
    // 删除连线
    graphRef.current?.removeCell(relationId);
  };

  return {
    initGraph,
    graph,
    graphRef,
    project,
    setProject,
    addTable,
    updateTable,
    deleteTable,
    addTopicArea,
    updateTopicArea,
    deleteTopicArea,
    addRemark,
    updateRemark,
    deleteRemark,
    addRelation,
    updateRelation,
    deleteRelation,
    canRedo,
    canUndo,
    onRedo,
    onUndo,
  };
}
